/*****************************************************************************
 * Author:   Valient Gough <vgough@pobox.com>
 *
 *****************************************************************************
 * Copyright (c) 2002-2004, Valient Gough
 *
 * This library is free software; you can distribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL), as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the LGPL in the file COPYING for more
 * details.
 *
 */
		                                                                                

#ifndef _rlog_incl_
#define _rlog_incl_

/*! @file rlog.h
  @brief Defines macros for debug, warning, and error messages.

*/

/*! @def RLOG_COMPONENT
    @brief Specifies build-time component, eg -DRLOG_COMPONENT="component-name"

    Define RLOG_COMPONENT as the name of the component being built.
    For example, as a compile flag,  -DRLOG_COMPONENT="component-name"

    If RLOG_COMPONENT is not specified, then it will be defined as "[unknown]"
*/
#ifndef RLOG_COMPONENT
#warning RLOG_COMPONENT not defined - setting to UNKNOWN
#define RLOG_COMPONENT "[unknown]"
#endif

/*
    Defined by configure if our compiler allows us to specify printf attributes
    on a function pointer..  Newer versions of GCC allow this, but older ones
    do not..
*/
#define HAVE_PRINTF_FP 1

/*
    We use __printf__ attribute to allow gcc to inspect printf style arguments
    and give warnings if the rDebug(), rWarning(), etc macros are misused.

    We use __builtin_expect on GCC 2.96 and above to allow optimization of
    publication activation check.  We tell the compiler that the branch is
    unlikely to occur, allowing GCC to push unecessary code out of the main
    path.
*/
#ifdef __GNUC__

# define PRINTF(FMT,X) __attribute__ (( __format__ ( __printf__, FMT, X)))
# define HAVE_PRINTF_ATTR 1

# if !defined( __builtin_expect )
#  define __builtin_expect(x,expected_value) (x)
# endif

# define   likely(x)  __builtin_expect((x),1)
# define unlikely(x)  __builtin_expect((x),0)

#else

// Not using the gcc compiler, make the macros do nothing..  They are
// documented as the last instance of the macros..

/*! @def PRINTF(FMT,X)
  On GCC, this uses the compiler's __printf__ attribute to tell the compiler
  to treat certain arguments as printf formating options, which allows it to
  print warnings at compile time if the arguments do not match the format
  string.
  @internal
*/
# define PRINTF(FMT,X)
# define HAVE_PRINTF_ATTR 0
/*!
*/
# define   likely(x)  (x)
/*! @def unlikely(x)
  Starting with GCC 2.96, we can tell the compiler that an if condition is
  likely or unlikely to occur, which allows the compiler to optimize for the
  normal case.
  @internal
*/
# define unlikely(x)  (x)

#endif

#if HAVE_PRINTF_FP
# define PRINTF_FP(FMT,X) PRINTF(FMT,X)
#else
# define PRINTF_FP(FMT,X)
#endif

#define CONCAT2(A,B) A##B
#define CONCAT(A,B) CONCAT2(A,B)
#define STR(X) #X

// Use somewhat unique names (doesn't matter if they aren't as they are used in
// a private context, so the compiler will make them unique if it must)
#define LOGID CONCAT(_rL_, __LINE__)

// may be useful for checking from configuration files
#define CURRENT_RLOG_VERSION 20040219
extern "C" int RLogVersion();

namespace rlog
{

    class RLogChannel;
    class RLogPublisher;

    /*! @enum LogLevel
      Logging level definitions.
    */
    enum LogLevel
    {
	Log_Undef    =0, //!< undefined level
	Log_Critical,    //!< critical conditions
	Log_Error,       //!< error conditions
	Log_Warning,     //!< warning conditions
	Log_Notice,      //!< normal, but significant, condition
	Log_Info,        //!< informational
	Log_Debug        //!< debug level messages
    };

    /*! @fn RLogInit(int &argc, char **argv)
      @brief Initializer for external rlog modules.

      This should be called at the beginning of a program which uses RLog.  If
      it isn't called, the logging should still work anyway although some
      external features may not.

      This allows command line arguments to be passed to any rlog modules which
      might want to examine them.
    */
    void RLogInit(int &argc, char **argv);


    // Get channel with a particular component name
    RLogChannel *GetComponentChannel(const char *component, 
	                             const char *path, 
				     LogLevel level = Log_Undef);

    // the global channel receives messages for all components
    RLogChannel *GetGlobalChannel( const char *path,
				   LogLevel level = Log_Undef);

    /*! @def DEF_CHANNEL( const char *path, LogLevel level )
      @brief Returns pointer to RLogChannel struct for the given path
      @param path The hierarchical path to the channel.  Elements in the path
      are separated by '/'.

      DEF_CHANNEL gets an existing (or defines a new) log type.  For example
      "debug", "warning", "error" are predefined types.  You might define
      completely new types, like "timing", or perhaps sub-types like
      "debug/timing/foo", depending on your needs.

      Reporting paths do not need to be unique within a project (or even a
      file).

      Channels form a hierarchy.  If one subscribes to "debug", then you also
      get messages posted to more specific types such as "debug/foo".  But if
      you subscribe to a more specific type, such as "debug/foo", then you will
      not receive more general messages such as to "debug".

      Example:
      @code
      #include <rlog/rlog.h>
      #include <rlog/RLogChannel.h>

      static RLogChannel *MyChannel = DEF_CHANNEL("me/mine/allmine",Log_Info);

      func()
      {
          rLog( MyChannel, "this is being sent to my own channel" );
	  rLog( MyChannel, "%s %s", "hello", "world" );
      }

      main()
      {
          // log all messages to the "me" channel to stderr
          StdioNode stdLog( STDERR_FILENO );
          stdLog.subscribeTo( RLOG_CHANNEL ("me") );

          func();
      }
      @endcode

      @see test.cpp
     */
#define DEF_CHANNEL(path,level) RLOG_CHANNEL_IMPL(RLOG_COMPONENT, path, level)
#define RLOG_CHANNEL(path) RLOG_CHANNEL_IMPL(RLOG_COMPONENT, path, Log_Undef)
#define RLOG_CHANNEL_IMPL(COMPONENT,path,level) \
    rlog::GetComponentChannel(STR(COMPONENT),path,level)

    /*
	Pre-defined channels, 
	"debug", "warning", and "error".

	You can of course defined sub-channels based on the predefined types,
	such as "debug/level1", or whatever.
    */
    extern RLogChannel *_RLDebugChannel;
    extern RLogChannel *_RLWarningChannel;
    extern RLogChannel *_RLErrorChannel;

    /*! @struct PublishLoc <rlog/rlog.h>
      @brief Internal RLog structure - static struct for each rLog() statement

      @internal
      Structure created for each log location to keep track of logging state
      and associated data.

      Only static members are initialized at build time, which is why
      RLogChannel is passed as argument.  Otherwise entire structure will have
      to be initialized at run-time which adds extra code and a guard variable
      for the struct.
    */
    struct PublishLoc
    {
	// If the compiler supports printf attribute specification on function
	// pointers, we'll use it here so that the compiler knows to check for
	// proper printf formatting.  If it doesn't support it, then we'll
	// force the check by inserting a bogus inline function..
	//! function to call when we reach the log statement.
	void (*publish)(PublishLoc *, RLogChannel *, const char *format, ... )
	    PRINTF_FP(3,4);
	RLogPublisher *pub;
	const char *component;
	const char *fileName;
	const char *functionName;
	int lineNum;
	RLogChannel *channel;
    };

    /*! @fn RLog_Register
      @brief Internal RLog function - registers a log statement. 
      
      This is used by the rDebug(), rWarning(), rError() macros.
      @internal
     */
    void RLog_Register(PublishLoc *loc, RLogChannel *, 
	               const char *format, ... ) PRINTF(3,4);

    /*! @fn rAssertFailed
      @brief Internal RLog function - helper function for rAssert macro.
    
      @internal
    */
    void rAssertFailed( const char *component, const char *file, 
	                const char *function, int line, 
	                const char *conditionStr );

    // if we don't have printf attributes on function pointers, but we do have
    // printf attributes, then make a bogus check function..
#if !HAVE_PRINTF_FP && HAVE_PRINTF_ATTR
    void __checkArgs(const char *, ... ) PRINTF(1,2);

    inline void __checkArgs(const char *, ...)
    { }
#endif
}

/*! @def _rMessageDef
  Defines a static RLogPublisher and points it to the registration function for
  the first call.
  @internal
*/
#define _rMessageDef(ID, COMPONENT) \
  static rlog::PublishLoc ID ={&rlog::RLog_Register, 0, STR(COMPONENT), \
      __FILE__, __FUNCTION__, __LINE__, 0};

/*! @def _rMessageCall
  Checks if the RLogPublisher is enabled and publishes the message if so.
  @internal
*/
#if HAVE_PRINTF_FP || !HAVE_PRINTF_ATTR
#define _rMessageCall(ID, CHANNEL, ...) \
  if(unlikely(ID.publish!=0)) (*ID.publish)( &ID, CHANNEL, ##__VA_ARGS__ );
#else
#define _rMessageCall(ID, CHANNEL, ...) \
  if(unlikely(ID.publish!=0))  \
  { \
    (*ID.publish)( &ID, CHANNEL, ##__VA_ARGS__ ); \
    rlog::__checkArgs( __VA_ARGS__ ); \
  }
#endif

/*! @def _rMessage(ID, CHANNEL, ... )
 
  Combines the publisher definition (_rMessageDef) and invokation
  (_rMessageCall)

  enclose in do{}while(0) to insure that it acts as a single statement even if
  placed in various if/else constructs..
  @internal
*/
#define _rMessage(ID, CHANNEL, ... ) \
  do { _rMessageDef(ID, RLOG_COMPONENT) \
       _rMessageCall(ID, CHANNEL, ##__VA_ARGS__ ) } while(0)

/*! @addtogroup RLogMacros
  These macros are the primary interface for logging messages:
  - rDebug(format, ...)
  - rWarning(format, ...)
  - rError(format, ...)
  - rLog(channel, format, ...)
  - rAssert( condition )
  @{
*/

/*! @def rDebug(format, ...)
    @brief Log a message to the "debug" channel.  Takes printf style arguments.

    Format is ala printf, eg:
    @code
    rDebug("I'm sorry %s, I can't do %s", name, request);
    @endcode

    When using a recent GNU compiler, it should automatically detect format
    string / argument mismatch just like it would with printf.

    Note that unless there are subscribers to this message, it will do nothing.
*/
#define rDebug(...) \
  _rMessage( LOGID, rlog::_RLDebugChannel, ##__VA_ARGS__ )

/*! @def rWarning(format, ...)
    @brief Log a message to the "warning" channel.  Takes printf style
    arguments.

    Output a warning message - meant to indicate that something doesn't seem
    right.

    Format is ala printf, eg:
    @code
    rWarning("passed %i, expected %i, continuing", foo, bar);
    @endcode

    When using a recent GNU compiler, it should automatically detect format
    string / argument mismatch just like it would with printf.

    Note that unless there are subscribers to this message, it will do nothing.
*/
#define rWarning(...) \
  _rMessage( LOGID, rlog::_RLWarningChannel, ##__VA_ARGS__ )

/*! @def rError(...)
    @brief Log a message to the "error" channel. Takes printf style arguments.

    An error indicates that something has definately gone wrong.

    Format is ala printf, eg:
    @code
    rError("bad input %s, aborting request", input);
    @endcode

    When using a recent GNU compiler, it should automatically detect format
    string / argument mismatch just like it would with printf.

    Note that unless there are subscribers to this message, it will do nothing.
*/
#define rError(...) \
  _rMessage( LOGID, rlog::_RLErrorChannel, ##__VA_ARGS__ )

/*! @def rLog(channel,format,...)
    @brief Log a message to a user defined channel. Takes a channel and printf
    style arguments.

    An error indicates that something has definately gone wrong.

    Format is ala printf, eg:
    @code
    static RLogChannel * MyChannel = RLOG_CHANNEL( "debug/mine" );
    rLog(MyChannel, "happy happy, joy joy");
    @endcode

    When using a recent GNU compiler, it should automatically detect format
    string / argument mismatch just like it would with printf.

    Note that unless there are subscribers to this message, it will do nothing.
*/
#define rLog(channel, ...) \
  _rMessage( LOGID, channel, ##__VA_ARGS__ )

#define _rAssertFailed(COMPONENT, COND) \
    rlog::rAssertFailed(STR(COMPONENT),__FILE__,__FUNCTION__,__LINE__, COND)

/*! @def rAssert( cond )
    @brief Assert condition - throws error if @a cond evaluates to false.

    Assert error condition.  Displays error message if condition does not
    evaluate to TRUE.

    We throw an error from rAssertFailed.  It isn't done inline so that we
    don't have to include the STL exception code here and bloat callers that
    don't use it or don't care.
*/
#define rAssert( cond ) \
    do { \
	if( unlikely((cond) == false) ) \
	{ rError( "Assert failed: " STR(cond) ); \
	  _rAssertFailed(RLOG_COMPONENT, STR(cond)); \
	} \
    } while(0)

/*! @def rAssertSilent( cond )
    @brief Assert condition - throws error if @a cond evaluates to false, but
    does not display error message.

    Assert error condition.  Similar to rAssert except that it does not display
    an error.
*/
#define rAssertSilent( cond ) \
    do { \
	if( unlikely((cond) == false) ) \
	{ _rAssertFailed(RLOG_COMPONENT, STR(cond)); } \
    } while(0)


#define _ERROR_IMPL(COMPONENT, MSG) \
    rlog::Error( STR(COMPONENT), __FILE__, __FUNCTION__, __LINE__, MSG )

#define ERROR( MSG ) _ERROR_IMPL( RLOG_COMPONENT, MSG )
    

/*! @}
*/

#endif

